<?php

namespace Drupal\county\Plugin\Field\FieldFormatter;

use Drupal\Core\Database\Database;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\FormatterBase;
use Drupal\Core\Form\FormStateInterface;

/**
 * Base class for county field formatters.
 */
abstract class Base extends FormatterBase {

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return [
      'display_all_text' => TRUE,
      'display_all_except' => FALSE,
      'limit_count' => 0,
    ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   *
   * When called from field_ui the field definition is a FieldConfig object and
   * is that from the field in question.
   *
   * When called from views_ui the field definition is a BaseFieldDefinition, is
   * not related to the field in question, and only has default values for its
   * settings. As usual, a base field has no bundle property.
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $form = parent::settingsForm($form, $form_state);

    // Build form elements.
    $form['display_all_text'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Replace all values'),
      '#description' => $this->t('If field value includes all values for the administrative area, then replace list of individual values with the text "All counties".'),
      '#default_value' => $this->getSetting('display_all_text'),
    ];

    $form['display_all_except'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Display missing values'),
      '#description' => $this->t('Display all values for the administrative area not included in the field value. Text will read "All counties except ...".<br />This setting only applies if the number of missing values is less than the number of values.<br />If no missing values, then text will read "All counties".<br />The exception list will be limited to number of values indicated below.'),
      '#default_value' => $this->getSetting('display_all_except'),
    ];

    $form['limit_count'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Number of values'),
      '#description' => $this->t('Limit display to this number of values based on alphabetical sort order. If zero, then display all values.'),
      '#default_value' => $this->getSetting('limit_count'),
      '#size' => 5,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $settings = $this->getSettings();

    $summary = [];
    $summary[] = $settings['display_all_text'] ? $this->t('Replace all values with "All counties".') : NULL;
    $summary[] = $settings['display_all_except'] ? $this->t('Display "All counties except &lt;list&gt;" instead of each value.') : NULL;
    $summary[] = $settings['limit_count'] ? $this->t('Limit display to first @limit_count values based on alpha order.', ['@limit_count' => $settings['limit_count']]) : NULL;
    $summary = array_values(array_filter($summary));
    $summary += [$this->t('Display each value.')];

    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {
    $elements = [];
    if (!$items->count()) {
      return $elements;
    }

    // Also $items->getSettings();
    $settings = $this->getFieldSettings();
    if (empty($settings['admin_area_field'])) {
      return $elements;
    }
    $admin_area_field = $settings['admin_area_field'];
    $entity = $items->getEntity();
    if (empty($entity->{$admin_area_field})) {
      return $elements;
    }

    $admin_area = $entity->{$admin_area_field}->value;
    $options = $this->countyDisplayValue($admin_area);

    $display_missing = $display_limit = FALSE;
    $settings = $this->getSettings();

    // An array with structure [delta][value => xxx, _attributes => []].
    $values = $items->getValue();

    $diff = [];
    if (!empty($settings['display_all_text']) || !empty($settings['display_all_except'])) {
      $items_transposed = $this->transposeOptionsArray($values);
      // @todo Do array_intersect_key instead of unset?
      unset($items_transposed['_attributes']);
      $items_transposed = array_flip($items_transposed['value']) + array_flip(['_none', 'all']);
      $diff = array_diff_key($options, $items_transposed);
      if (!empty($settings['display_all_text'])) {
        if (empty($diff)) {
          $elements[0] = ['#markup' => $this->t('All counties')];
          return $elements;
        }
      }
      if (!empty($settings['display_all_except'])) {
        if ($items->count() > count($diff)) {
          // Display the missing values.
          $display_missing = TRUE;
          $values = $this->transposeOptionsArray(['value' => array_keys($diff)]);
        }
      }
    }

    $count = $limit = 0;
    if (!empty($settings['limit_count'])) {
      $limit = (int) $settings['limit_count'];
      if ($limit && $limit < count($values)) {
        $display_limit = TRUE;
        $count = count($values);
        $values = array_slice($values, 0, $limit);
      }
    }

    $elements = $this->viewValues($values, $admin_area, $settings);

    // @todo The next two items do not work well in a view set to display all
    // county values on one line with a simple separator (including a comma or
    // similar character). The separator added before the first item and after
    // the last makes for incorrect punctuation. Try to detect whether this is a
    // view with that setting and concatenate these strings to the first and
    // last items, respectively.
    if ($display_missing) {
      $begin = ['#markup' => count($diff) ? $this->t('All counties except') : $this->t('All counties')];
      array_unshift($elements, $begin);
    }
    if ($display_limit) {
      $elements[] = ['#markup' => $this->t('(first @limit of @count values)', ['@limit' => $limit, '@count' => $count])];
    }

    return $elements;
  }

  /**
   * Returns the render array for the massaged field items.
   *
   * @param array $items
   *   The massaged field items (or the all except items).
   * @param string $admin_area
   *   The administrative area.
   * @param array $settings
   *   The formatter settings (updated for missing default values).
   *
   * @return array
   *   The textual output generated as a render array.
   */
  protected function viewValues(array $items, $admin_area, array $settings) {
    return [];
  }

  /**
   * Returns an associative array of county display values or one display value.
   *
   * @param string $admin_area
   *   The administrative area value.
   * @param string $county
   *   (optional) The county FIPS value.
   *
   * @return array|string
   *   An associative array of county display values or one display value.
   */
  protected function countyDisplayValue($admin_area, $county = NULL) {
    $values = &drupal_static(__FUNCTION__, []);

    if (!isset($values[$admin_area])) {
      // Retrieve values from database.
      $args = [':state' => $admin_area];
    $options = Database::getConnection()->query('SELECT CONCAT(fips_state, fips_county) AS fips, county FROM {county_fips} WHERE state = :state', $args)->fetchAllKeyed();
      if (!$options) {
        $options = [];
      }
      // Add default values.
      // @todo Need this?
      $options['_none'] = 'None';
      $options['all'] = 'All counties';
      $values[$admin_area] = $options;
      // No harm to set this multiple times (potentially but not likely).
      $values['_none'] = ['_none' => 'None'];
    }

    if (isset($county)) {
      return isset($values[$admin_area][$county]) ? $values[$admin_area][$county] : 'None';
    }
    return $values[$admin_area];
  }

  // Copied from 7x options_array_transpose.
  // @todo What is 8x equivalent?

  /**
   * Manipulates a 2D array to reverse rows and columns.
   *
   * The default data storage for fields is delta first, column names second.
   * This is sometimes inconvenient for field modules, so this function can be
   * used to present the data in an alternate format.
   *
   * @param $array
   *   The array to be transposed. It must be at least two-dimensional, and
   *   the subarrays must all have the same keys or behavior is undefined.
   * @return
   *   The transposed array.
   */
  function transposeOptionsArray($array) {
    $result = [];
    if (is_array($array)) {
      foreach ($array as $key1 => $value1) {
        if (is_array($value1)) {
          foreach ($value1 as $key2 => $value2) {
            if (!isset($result[$key2])) {
              $result[$key2] = [];
            }
            $result[$key2][$key1] = $value2;
          }
        }
      }
    }
    return $result;
  }

}
