<?php

namespace Drupal\county\Plugin\Field\FieldWidget;

use Drupal\county\Plugin\AdminAreaTrait;
use Drupal\Component\Utility\Html;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\PrependCommand;
use Drupal\Core\Ajax\ReplaceCommand;
use Drupal\Core\Database\Database;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\Plugin\Field\FieldWidget\OptionsSelectWidget;
use Drupal\Core\Field\WidgetBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element\StatusMessages;

/**
 * Plugin implementation of the 'county_select' widget.
 *
 * @FieldWidget(
 *   id = "county_select",
 *   label = @Translation("County select"),
 *   field_types = {
 *     "county"
 *   },
 *   multiple_values = TRUE
 * )
 */
class Select extends OptionsSelectWidget {

  use AdminAreaTrait;

  /**
   * The administrative area.
   *
   * @var string
   */
  protected $adminArea;

  /*
   * @todo The OptionsWidgetBase class uses but does not define: required,
   *   multiple, has_value, options.
   */

  /**
   * Whether the field is required.
   *
   * @var bool
   */
  protected $required;

  /**
   * Whether the field allows for multiple values.
   *
   * @var bool
   */
  protected $multiple;

  /**
   * Whether the field has values in storage.
   *
   * @var bool
   */
  protected $has_value;

  /**
   * The list of allowed values.
   *
   * @var array
   */
  protected $options;

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return [
      'display_type' => 'long',
      'append_fips' => FALSE,
      'select_all' => FALSE,
    ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $settings = $this->getSettings();

    foreach (['long', 'short', 'key'] as $key) {
      $display_types[$key] = t(ucfirst($key));
    }

    $element['display_type'] = [
      '#type' => 'select',
      '#title' => t('Display type'),
      '#default_value' => $settings['display_type'],
      '#options' => $display_types,
      '#description' => t('Select the type of value to display on edit form.'),
      '#required' => TRUE,
    ];
    $element['append_fips'] = [
      '#type' => 'checkbox',
      '#title' => t('Append FIPS code'),
      '#default_value' => $settings['append_fips'],
      '#description' => t('If checked, then append FIPS code to display value.'),
      '#element_validate' => [[get_class($this), 'validateAppendFips']],
      '#states' => [
        'visible' => [
          ':input[name="settings[display_type]"]' => ['!value' => 'key'],
        ],
      ],
      '#field_name' => $this->fieldDefinition->getName(),
    ];
    $element['select_all'] = [
      '#type' => 'checkbox',
      '#title' => t('Select all'),
      '#default_value' => $settings['select_all'],
      '#description' => t('Add option to select all counties to the county list.'),
    ];

    return $element;
  }

  /**
   * Form validation handler for append_fips element.
   *
   * @param array $element
   *   The form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   */
  public static function validateAppendFips(array $element, FormStateInterface $form_state) {
    // Ensure append_fips is false if display_type is key.
    $field_name = $element['#field_name'];
    $values = &$form_state->getValues()['fields'][$field_name]['settings_edit_form']['settings'];
    $values['append_fips'] = $values['display_type'] == 'key' ? 0 : $values['append_fips'];
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $settings = $this->getSettings();

    $summary = [];
    $summary[] = t('Use @display_type display', ['@display_type' => $settings['display_type']]);
    $summary[] = $settings['append_fips'] ? t('Append FIPS code') : NULL;
    $summary[] = $settings['select_all'] ? t('Select all option') : NULL;
    $summary = array_values(array_filter($summary));

    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function formElement(FieldItemListInterface $items, $delta, array $element, array &$form, FormStateInterface $form_state) {
    $this->adminArea = $this->getAdminArea($items, $element, $form, $form_state);
    $element = parent::formElement($items, $delta, $element, $form, $form_state);

    // Display up to 12 rows, not the 4 used by OptionsSelectWidget.
    $element['#size'] = $element['#multiple'] ? min(12, count($element['#options'])) : 0;

    // Store id for later use on the wrapper element.
    $parents_key = implode('-', $element['#field_parents']);
    $county_field_name = $this->fieldDefinition->getName();
    $element['#county_id'] = static::getWrapperId($parents_key, $county_field_name);

    // To handle multiple county elements on a form (but not necessarily
    // across multiple forms on a page request), tied to the same or different
    // admin_area elements, store an array keyed by the admin_area_field_name
    // so that refreshCountyWidget() can loop on the county fields tied to
    // the triggering admin_area element.
    $settings = $this->getFieldSettings();
    $element['#admin_area_field_name'] = $admin_area_field_name = $settings['admin_area_field'];
    $langcode = $form_state->get('langcode');
    $property = ['county_field', $parents_key, $admin_area_field_name, $county_field_name];
    $form_state->set($property, $langcode);

    return $element;

    // If form is being built in response to AJAX, then clear inputs.
    // This has no effect as form api resets $form_state inputs after form build
    // and before form process.
    $inputs = &$form_state->getUserInput();

    // Use strpos() because a plain text field has
    // _triggering_element_name => field_name[0][value].
    if (!empty($inputs['_triggering_element_name']) && strpos($inputs['_triggering_element_name'], $admin_area_field_name) !== FALSE) {
      if (!empty($inputs[$county_field_name])) {
        unset($inputs[$county_field_name]);
        $form_state->setUserInput($inputs);
      }
    }

    return $element;
  }

  /**
   * Form validation handler for county widget element.
   *
   * @see Drupal/Core/Field/FieldItemList::defaultValuesForm{,Validate,Submit}
   */
  public static function validateElement(array $element, FormStateInterface $form_state) {
    // If form is being validated on AJAX request, then clear inputs.
    $admin_area_field_name = $element['#admin_area_field_name'];
    $county_field_name = $element['#field_name'];
    $inputs = &$form_state->getUserInput();

    // @todo See conditions used in ManyToOne::validateElement().
    if ($triggering_element = $form_state->getTriggeringElement()) {
      if (isset($triggering_element['#field_name']) && $triggering_element['#field_name'] == $admin_area_field_name) {
        if (!empty($inputs[$county_field_name])) {
          // The admin_area element was changed triggering an AJAX callback,
          // and the county element has values (based on the original value of
          // the admin_area element). The county element is rebuilt based on the
          // new admin_area value, which automatically produces an inconsistent
          // state between its submitted values and the new options.
          // To avoid bogus errors in FormValidator::performRequiredValidation()
          // clear the county inputs.
          // @todo For this to take effect requires commit of
          // https://www.drupal.org/project/drupal/issues/3137947.
          unset($inputs[$county_field_name]);
          $form_state->setUserInput($inputs);

          // Clear the form errors (and, if possible, the log entries).
          // @todo Remove next if block after commit of
          // https://www.drupal.org/project/drupal/issues/3137947.
          $errors = $form_state->getErrors();
          if (isset($errors[$county_field_name]) && strpos($errors[$county_field_name], 'An illegal choice has been detected') !== FALSE) {
            // Remove error message.
            $form_state->clearErrors();
            // @todo The logger writes error to log file. Can we prevent this?
//             LoggerChannelTrait::getLogger();
          }
        }
        return;
      }
    }

    // @see OptionsWidgetBase::validateElement which does first check.
    // Also does this widget need the value manipulation done in parent routine?
    if ($element['#multiple'] && is_array($element['#value'])) {
      if (count($element['#value']) == 1) {
        // @todo Update this comment for 8x.
        // This manipulation is needed because this module always adds '_none'
        // item. The options module only adds this item when multiple and not
        // required, in which case it does not care whether '_none' is selected
        // and removes it in _options_form_to_storage().
        $element['#value'] = current($element['#value']);
        if ($element['#required'] && $element['#value'] == '_none') {
          $form_state->setError($element, t('!name field is required.', ['!name' => $element['#title']]));
        }
      }
      elseif (in_array('all', $element['#value'])) {
        $form_state->setError($element, t('!name field may either be all counties or individual counties.', ['!name' => $element['#title']]));
      }
    }

    parent::validateElement($element, $form_state);
  }

  /**
   * Returns a fixed HTML ID for the county wrapper element.
   *
   * @param string $parents_key
   *   The array key based on parents array.
   * @param string $county_field_name
   *   The county field name.
   *
   * @return string
   *   The HTML ID for the county wrapper element.
   */
   public static function getWrapperId($parents_key, $county_field_name) {
    // Allow for $parents_key to be blank and omit it from $parts.
    $parts = array_filter(['edit', $parents_key, str_replace('_', '-', $county_field_name), 'wrapper']);
    return implode('-', $parts);
  }

  /**
   * Returns the administrative area value.
   *
   * This is either the last submitted value, the entity value, or NULL.
   *
   * @return string
   *   The administrative area value.
   */
  protected function getAdminArea(FieldItemListInterface $items, array $element, array &$form, FormStateInterface $form_state) {
    if ($this->isDefaultValueWidget($form_state)) {
      return '_none';
    }

    $settings = $this->fieldDefinition->getSettings();
    $admin_area_field_name = $settings['admin_area_field'];

    $key_exists = NULL;
    $values = NULL;

    $inputs = $form_state->getUserInput();

    if (!empty($inputs)) {
      // Start with submitted values.

      // Input values for the entire form are always in $inputs except for
      // disabled elements. This is not true of $form_state->values if the
      // triggering element has a '#limit_validation_errors' property that
      // excludes some of the county elements. This is likely to occur with
      // multiple county elements on a form, or multiple elements with an AJAX
      // callback.

      // For a list_string field, the 'inputs' array ends at the 'field_name'
      // key so cast it to an array (equivalent to adding a delta key) to ensure
      // an array for consistency and as expected by code follwing this block.
      $path = array_merge($form['#parents'], [$admin_area_field_name]);
      $values = NestedArray::getValue($inputs, $path, $key_exists);
      $values = (array) $values;
    }

/*
    if (!$key_exists && isset($element['#entity'])) {
      // Fall back to entity values.
      $entity_type = $element['#entity_type'];
      $entity = $element['#entity'];
      $values = field_get_items($entity_type, $entity, $admin_area_field_name, $langcode);
    }
*/
    if (!$key_exists) {
      // Fall back to entity values.
      // @todo This works with entity form but not with view filter.
      $entity = $items->getEntity();
      $values = $entity->get($admin_area_field_name)->getValue();
    }

    // Shift array values until we end up with a string. Depending on the field
    // type and cardinality (along with value source), this implies 1-2 shift
    // operations.
    // For values from the entity, the array is [field_name][delta][column].
    // From inputs, the array is:
    // - [field_name][delta] for list_string (with [delta] added by cast)
    // - [field_name][delta][column, {_weight}] for string.
    // A _weight key is present in delta array if cardinality > 1.
    $values = is_array($values) ? array_shift($values) : '_none';
    $value = is_array($values) ? array_shift($values) : $values;

    return $value;
  }

  /**
   * Returns the array of options for the widget.
   *
   * @param \Drupal\Core\Entity\FieldableEntityInterface $entity
   *   The entity for which to return options.
   *
   * @return array
   *   The array of options for the widget.
   */
  protected function getOptions(FieldableEntityInterface $entity) {
    if (!isset($this->options)) {
      $options = [];
      if ($this->adminArea != '_none') {
        // @todo Check the '#empty_value' property of administrative area field.
        $options = static::getCountyOptions($this->adminArea, $this->getSettings());
      }

      // Add an empty option if the widget needs one.
      if ($empty_label = $this->getEmptyLabel()) {
        $options = ['_none' => $empty_label] + $options;
      }
      $this->options = $options;
    }
    return $this->options;
  }

  /**
   * Determines selected options from the incoming field values.
   *
   * @todo Could this remove stale items and avoid code in validateElement?
   *
   * @param \Drupal\Core\Field\FieldItemListInterface $items
   *   The field values.
   *
   * @return array
   *   The array of corresponding selected options.
   */
  protected function getSelectedOptions__DEP(FieldItemListInterface $items) {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  protected function sanitizeLabel(&$label) {
    // Select form inputs allow unencoded HTML entities, but no HTML tags.
    $label = Html::decodeEntities(strip_tags($label));
  }

  /**
   * {@inheritdoc}
   */
  protected function getEmptyLabel() {
    if ($this->multiple) {
      // Multiple select: add a 'none' option for non-required fields.
      if (!$this->required) {
        return t('- None -');
      }
    }
    else {
      // Single select: add a 'none' option for non-required fields,
      // and a 'select a value' option for required fields that do not come
      // with a value selected.
      if (!$this->required) {
        return t('- None -');
      }
      if (!$this->has_value) {
        return t('- Select a value -');
      }
    }
  }

  /**
   * Adds AJAX callback to administrative area form element.
   *
   * This does the following:
   * - sets fixed ID on the county field wrapper DIV (needed for AJAX callback)
   * - adds AJAX handling to a change of the administrative area field.
   *
   * In 7x, preferred hook_field_attach_form() to:
   *   hook_form_alter() as it is called:
   *   - only in the context of a form with a field
   *   an after_build callback as it is called:
   *   - before the form is cached (this is the retrieved form)
   *   - as drupal_process_form() caches the $unprocessed_form
   *   - so the '#ajax' stuff is present when called by ajax_form_callback()
   * In 8x, only appear to have hook_form_alter().
   *
   * @todo Does the ajax api support adding the ajax stuff to the county element
   *   but having it be triggered by the state element? If so that would greatly
   *   simplify things.
   */
  public static function fieldAttachForm(array &$form, FormStateInterface $form_state, $form_id) {
    if (empty($form_state->get('county_field')) || $form_state->getFormObject() instanceof \Drupal\field_ui\Form\FieldConfigEditForm) {
      // No county field or form is used for default value widget.
      return;
    }
    if (!method_exists($form_state->getFormObject(), 'getEntity')) {
      // Most likely a view exposed form.
      return;
    }

    $entity = $form_state->getFormObject()->getEntity();

    foreach ($form_state->get('county_field') as $parents_key => $admin_area_field_names) {
      foreach ($admin_area_field_names as $admin_area_field_name => $county_field_names) {
        foreach ($county_field_names as $county_field_name => $langcode) {
          if (empty($form[$county_field_name])) {
            continue;
          }
          $element = &$form[$county_field_name];
          if (!empty($element['#county_processed'])) {
            continue;
          }
          $element['#county_processed'] = 1;
          // Fix the HTML ID for the Ajax callback.
          // @todo Append the entity ID in case of multiple forms on page.
          $element['#id'] = $element['widget']['#county_id'];
        }

        // @todo Is this element always present at this key?
        if (!isset($form[$admin_area_field_name])) {
          // This hook may be invoked more than once on a form where $form is
          // only a subset of the entire form and omits admin_area_field_name,
          // but $form_state includes the 'county_field' value.
          continue;
        }
        $admin_area_element = &$form[$admin_area_field_name];
        if (!empty($admin_area_element['#county_processed'])) {
          continue;
        }
        $admin_area_element['#county_processed'] = 1;

        // @todo Entity reference widget adds fallback for non-JS submit using a
        // hidden submit button.
        $parents = $form['#parents'];
        $ajax = [
          // Only validate the administrative area field.
          // This eventually makes it the only value in $form_state->values.
          '#limit_validation_errors' => [
            array_merge($parents, [$admin_area_field_name]),
          ],
          '#ajax' => [
            'callback' => [get_class(), 'refreshCountyWidget'],
            // @todo Add this to $settings in ajax ReplaceCommand()?
            'effect' => 'fade',
            'event' => 'change',
          ],
          // Add key for multiple value form; it is not at this depth.
          '#field_name' => $admin_area_field_name,
        ];

        $admin_area_field = $entity->getFieldDefinition($admin_area_field_name);
        $admin_area_field_type = $admin_area_field->getType();
        if ($admin_area_field_type == 'list_string') {
          // @todo Update this comment for 8x.
          // A list_string field uses the options module for its widgets; the latter sets
          // ['behaviors']['multiple values'] = FIELD_BEHAVIOR_CUSTOM. This property is
          // checked in field_default_form() to determine whether to call
          // field_multiple_value_form() and include a delta key in the form element.
          // Because of this property, a list_string field does not have a delta key in
          // its field form element.
          // See options_field_widget_info().

          // Check for a delta key before adding this ajax item.
          $has_delta_key = isset($admin_area_element['widget']['#theme']) && $admin_area_element['widget']['#theme'] == 'field_multiple_value_form';
          if ($has_delta_key) {
            $admin_area_element['widget'][0] += $ajax;
          }
          else {
            $admin_area_element['widget'] += $ajax;
          }
        }
        else {
          // A simple text field has a 'value' key which has the '#type' property
          // and needs to have the '#ajax' property so ajax_pre_render_element()
          // will process the element.
          // @todo If we are to support other field types, then we may need to look
          // at the field columns and use the first column in the array.
          $admin_area_element['widget'][0]['value'] += $ajax;
        }
      }
    }
  }

  /**
   * AJAX callback: Change of administrative area field.
   */
  public static function refreshCountyWidget(array &$form, FormStateInterface $form_state) {
    $response = new AjaxResponse();
    $selector = '';

    $trigger = $form_state->getTriggeringElement();
    $admin_area_field_name = isset($trigger['#field_name']) ? $trigger['#field_name'] : 'MISSING';
    // Any items preceding the field name are the parent keys.
    $parents = $trigger['#parents'];
    $key = array_search($admin_area_field_name, $parents);
    $parents = $key ? array_slice($parents, 0, $key) : [];
    $parents_key = implode('-', $parents);

    foreach ($form_state->get('county_field')[$parents_key][$admin_area_field_name] as $county_field_name => $langcode) {
      if (!empty($form[$county_field_name])) {
        $county_element = $form[$county_field_name];
      }
      else {
        // The county element is nested in form array; find the parent keys.
        $entity = $form_state->getFormObject()->getEntity();
        $field_state = WidgetBase::getWidgetState($parents, $county_field_name, $form_state);
        $parents = $field_state['array_parents'];
        $county_element = NestedArray::getValue($form, $parents, $key_exists);
        if (!$key_exists) {
          // @todo Do something more in this case.
          return [];
        }
      }
      // Remove the '_weight' element inserted by field API if cardinality != 1.
      unset($county_element['_weight']);
      $html = \Drupal::service('renderer')->render($county_element);
      if ($html) {
        $selector = '#' . static::getWrapperId($parents_key, $county_field_name);
        $response->addCommand(new ReplaceCommand($selector, $html));
      }
    }
    if ($selector) {
      // Prepend the messages to the last selector.
      // @see https://api.drupal.org/api/drupal/core%21core.api.php/group/ajax/8.7.x
      // The example code is stale: drupal_render(... 'status_messages')
      // That code needs to go in the form builder code so that the render
      // pipeline does the lazy load thing.
      $response->addCommand(new PrependCommand($selector, StatusMessages::renderMessages()));
    }
    return $response;
  }

}
