<?php

namespace Drupal\county\Plugin\views\filter;

use Drupal\Core\Form\FormStateInterface;
use Drupal\views\Plugin\views\filter\StringFilter;
use Drupal\views\Views;

/**
 * Plugin implementation of the 'basic' county filter.
 *
 * Filters on either the county FIPS value or county text.
 *
 * @ingroup views_filter_handlers
 *
 * @ViewsFilter("county_basic")
 *
 * @todo Might also allow syntax like 'KS:Riley'.
 */
class Basic extends StringFilter {

  /**
   * {@inheritdoc}
   */
  protected function defineOptions() {
    $options = parent::defineOptions();

    $options['value_type'] = ['default' => 'fips'];

    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function buildOptionsForm(&$form, FormStateInterface $form_state) {
    parent::buildOptionsForm($form, $form_state);

    $form['value_type'] = [
      '#type' => 'radios',
      '#title' => t('Value type'),
      '#description' => t('Indicate whether each filter value is text (Macon County) or FIPS (20161).'),
      '#default_value' => $this->options['value_type'],
      '#options' => ['text' => 'County text', 'fips' => 'FIPS value'],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function operators() {
    $operators = parent::operators();
    unset($operators['word'], $operators['allwords'], $operators['shorterthan'], $operators['longerthan'], $operators['regular_expression']);
    return $operators;
  }

  /**
   * {@inheritdoc}
   */
  function query() {
    if ($this->options['value_type'] == 'fips') {
      // This join will return a row for each delta row in the field table
      // whose value satisfies the filter criteria. If need be, add distinct
      // to the query settings of the view.
      parent::query();
      return;
    }

    $this->ensureMyTable();
    $field = "$this->tableAlias.$this->realField";

    // Join the county field table with the county_fips table.
    // The former has fips value; the latter has this split into two columns.
    $configuration = [
      'type'       => 'LEFT',
      'left_table' => $this->tableAlias,
      'left_formula' => "SUBSTR($field, 1, 2)",
      'left_field' => '',
      'table'      => 'county_fips',
      'field'      => 'fips_state',
      'extra'      => "SUBSTR($field, 3, 3) = county_fips.fips_county",
    ];
    $join = Views::pluginManager('join')->createInstance('standard', $configuration);
    $join->adjusted = TRUE;
    $fips = $this->query->queueTable('county_fips', $this->relationship, $join/*, $alias*/);
    if ($fips != 'county_fips') {
      // Adjust table alias in extra join clause.
      $queue = &$this->query->getTableQueue();
      $queue[$fips]['join']->extra = str_replace('county_fips', $fips, $queue[$fips]['join']->extra);
    }

    $info = $this->operators();

    if (!empty($info[$this->operator]['method'])) {
      if ($info[$this->operator]['method'] != 'op_empty') {
        // Conditions involving NULL need to be applied against real_field; all
        // others against the fips county value.
        $field = "$fips.county";
      }
      $this->{$info[$this->operator]['method']}($field);
    }
  }

  /**
   * Adds where clause to query.
   *
   * Each of these should compare filter value to 'county' column in county_fips table.
   * Use the relationship from the query above.
   *
   * @param string $field
   *   The field name.
   */
  function op_multiple($field) {
    $this->query->add_where($this->options['group'], $field, $this->value, $this->operator);
  }
}
