<?php

namespace Drupal\county\Plugin\views\filter;

use Drupal\county\Plugin\AdminAreaTrait;
use Drupal\Core\Form\FormStateInterface;
use Drupal\views\Plugin\views\display\DisplayPluginBase;
use Drupal\views\Plugin\views\filter\ManyToOne as ViewsManyToOne;
use Drupal\views\ViewExecutable;

/**
 * Plugin implementation of the 'many_to_one' county filter.
 *
 * Filters on the county FIPS value while displaying county text in select.
 *
 * @ingroup views_filter_handlers
 *
 * @ViewsFilter("county_many_to_one")
 */
class ManyToOne extends ViewsManyToOne {

  use AdminAreaTrait;

  /**
   * The administrative area.
   *
   * @var string
   */
  protected $adminArea;

  /**
   * The administrative area field name.
   *
   * @var string
   */
  protected $adminAreaField;

  /**
   * The administrative area source (argument or filter).
   *
   * @var string
   */
  protected $adminAreaSource;

  /**
   * {@inheritdoc}
   */
  protected function defineOptions() {
    $options = parent::defineOptions();

    $options['admin_area_argument'] = ['default' => ''];
    $options['admin_area_filter'] = ['default' => ''];

    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function buildOptionsForm(&$form, FormStateInterface $form_state) {
    parent::buildOptionsForm($form, $form_state);

    // Gather data.
    // Create an array of configured filters.
    $argument_fields = $this->optionsList('argument');
    $filter_fields = $this->optionsList('filter');

    // Build form elements.
    // Allow for an argument [contextual filter] and a filter as basis for the
    // admin_area_field. Indicate which is preferred.
    $form['admin_area_argument'] = [
      '#type' => 'select',
      '#title' => t('State field [argument]'),
      '#default_value' => $this->options['admin_area_argument'],
      '#options' => $argument_fields,
      '#empty_value' => '_none',
      '#empty_option' => '- None -',
      '#description' => t('Select the state argument used to filter the county list.'),
    ];
    $form['admin_area_filter'] = [
      '#type' => 'select',
      '#title' => t('State field [filter]'),
      '#default_value' => $this->options['admin_area_filter'],
      '#options' => $filter_fields,
      '#empty_value' => '_none',
      '#empty_option' => '- None -',
      '#description' => t('Select the state filter used to filter the county list.'),
    ];
  }

  /**
   * Returns associative array of fields by handler type.
   */
  protected function optionsList($type) {
    $options = [];
    // @todo Possible to filter fields by type; retain string and list_string?
    foreach ($this->view->display_handler->getHandlers($type) as $id => $handler) {
      if ($id == $this->options['id']) {
        continue;
      }
      $options[$id] = isset($handler->definition['title short']) ? $handler->definition['title short'] : $handler->definition['title'];
    }
    return $options;
  }

  /**
   * {@inheritdoc}
   */
  function query() {
    if (is_array($this->value) && in_array('all', $this->value)) {
      // All counties are allowed; no query additions needed.
      return;
    }
    parent::query();
  }

  /**
   * {@inheritdoc}
   *
   * Allow for an argument [contextual filter] and a filter as basis for the
   * admin_area_field. Indicate which is preferred.
   *
   * If admin_area_field comes from an exposed filter, then would like to do the
   * usual county widget form changes with AJAX when building the exposed form
   * element, but can not as this form uses a GET method and core AJAX only
   * works with a POST method.
   *
   * @see views_handler_filter_in_operator::value_form().
   */
  protected function valueForm(&$form, FormStateInterface $form_state) {
    parent::valueForm($form, $form_state);

    $this->getAdminAreaSettings($form, $form_state);
    if ($this->adminAreaSource == 'filter') {
      $form['#parents'] = [];
      $element = &$form['value'];
      $element['#element_validate'] = [[get_class($this), 'validateElement']];

      /*
      // OMIT: views exposed form uses 'GET' method, core ajax requires 'POST'.
      // Copied from Select::formElement().
      $parents_key = '';
      $county_field_name = $this->options['field'];
      $element['#county_id'] = Select::getWrapperId($parents_key, $county_field_name);

      $admin_area_field_name = $this->adminAreaField;
      $langcode = $form_state->get('langcode');
      $property = ['county_field', $parents_key, $admin_area_field_name, $county_field_name];
      $form_state->set($property, $langcode);

      // Attach AJAX to state element change.

      // Copied from Select::fieldAttachForm().
      // @see views_form_views_exposed_form_alter()
      // The above sets access = false on the form_build_id which is required
      // by regular core ajax processing.
      if (!empty($element['#county_processed'])) {
        return;
      }
      $element['#county_processed'] = 1;
      // Fix the HTML ID for the Ajax callback.
      // @todo Append the entity ID in case of multiple forms on page.
      $element['#id'] = str_replace('-wrapper', '', $element['#county_id']);

      $admin_area_element = &$form[$admin_area_field_name];
      if (!empty($admin_area_element['#county_processed'])) {
        return; // continue;
      }
      $admin_area_element['#county_processed'] = 1;

      $parents = $form['#parents'];
      $ajax = [
        // Only validate the administrative area field.
        // This eventually makes it the only value in $form_state->values.
        '#limit_validation_errors' => [
          array_merge($parents, [$admin_area_field_name]),
        ],
        '#ajax' => [
          'callback' => ['Drupal\county\Plugin\Field\FieldWidget\Select', 'refreshCountyWidget'],
          'effect' => 'fade',
          'event' => 'change',
        ],
      ];
      $admin_area_element += $ajax;
      */
    }
  }

  /**
   * Sets the admin_area field, value, and source (argument or filter).
   */
  function getAdminAreaSettings($form, $form_state) {
    $admin_area_field = $admin_area_source = '';
    $admin_area = '_none';
    // Try the filter then the argument.
    // @todo Omit this filter from the list.
    if (!empty($this->options['admin_area_filter']) && $this->options['admin_area_filter'] != '_none') {
      $id = $this->options['admin_area_filter'];
      $filter_handler = $this->view->display_handler->getHandler('filter', $id);
      if (!empty($this->view->getExposedInput()[$id])) {
        $this->adminAreaSource = 'filter';
        $this->adminAreaField = $id;
        $this->adminArea = $this->view->getExposedInput()[$id];
        return;
      }
      elseif (!empty($filter_handler->value)) {
        // The value property is not set at this point, unless set in custom code.
        $this->adminAreaSource = 'filter';
        $this->adminAreaField = $id;
        $this->adminArea = current($filter_handler->value);
        return;
      }
    }
    if (!empty($this->options['admin_area_argument']) && $this->options['admin_area_argument'] != '_none') {
      $id = $this->options['admin_area_argument'];
      $argument_handler = $this->view->display_handler->getHandler('argument', $id);
      if (!empty($argument_handler) && !empty($argument_handler->get_value())) {
        $this->adminAreaSource = 'argument';
        $this->adminAreaField = $id;
        $this->adminArea = $argument_handler->get_value();
        return;
      }
    }

    $this->adminAreaField = $admin_area_field;
    $this->adminArea = $admin_area;
    $this->adminAreaSource = $admin_area_source;
  }

  /**
   * {@inheritdoc}
   *
   * @todo
   * This is called to build settings form in views_ui and for exposed form.
   * For the former case, the handler has no value.
   */
  public function getValueOptions() {
    if (isset($this->valueOptions)) {
      return $this->valueOptions;
    }
    $this->getAdminAreaSettings([], []);
    // Views adds an option with an 'All' key only if one value is allowed on
    // the filter. This module adds an 'all' key that is always a valid choice.
    $this->valueOptions = static::getCountyOptions($this->adminArea, ['display_type' => 'short', 'select_all' => TRUE]);
  }

  /**
   * {@inheritdoc}
   *
   * Form validation handler for views_exposed_form().
   *
   * This could also be accomplished by JS to clear the county element on change
   * of the state element. Neither approach provides the desired solution to
   * refresh the county options on state change.
   *
   * @see views_exposed_form_validate()
   */
  public static function validateElement(array &$element, FormStateInterface $form_state, &$complete_form) {
    // If none of the county values is in current county options, then clear the
    // form_state inputs for county element to avoid bogus errors from the
    // FormValidator::performRequiredValidation() on a select element.
    $county_field_name = $element['#name'];
    $inputs = &$form_state->getUserInput();
    if ($triggering_element = $form_state->getTriggeringElement()) {
      if (!empty($inputs[$county_field_name]) && !array_intersect_key($element['#options'], array_flip($inputs[$county_field_name]))) {
        // None of the inputs (or submitted values) are in the current options.
        // The options and inputs are out of sync implying the admin_area value
        // was changed and the exposed form submit is the first opportunity to
        // update the options. Clear the inputs and element values to avoid
        // bogus error message and log entry in performRequiredValidation(),
        // which by patch of core is executed after this method.
        // @todo For this to take effect requires commit of
        // https://www.drupal.org/project/drupal/issues/3137947.
        unset($inputs[$county_field_name]);
        $form_state->setUserInput($inputs);
        // Clear element value.
        // @todo This is not necessary on the field widget where the admin_area
        // is the triggering_element and it uses limit_validation_errors, which
        // by patch of core is respected in performRequiredValidation().
        unset($element['#value']);
      }
    }
  }
}
